/*  Copyright 2003-2017 Ciaran Anscomb
 *
 *  This file is part of XRoar.
 *
 *  XRoar is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  XRoar is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with XRoar.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "slist.h"
#include "xalloc.h"

#include "logging.h"
#include "xconfig.h"

static struct xconfig_option const *find_option(struct xconfig_option const *options,
		const char *opt) {
	int i;
	for (i = 0; options[i].type != XCONFIG_END; i++) {
		if (0 == strcmp(options[i].name, opt)) {
			return &options[i];
		}
	}
	return NULL;
}

static int lookup_enum(const char *name, struct xconfig_enum *list, int undef_value) {
	int i;
	for (i = 0; list[i].name; i++) {
		if (0 == strcmp(name, list[i].name)) {
			return list[i].value;
		}
	}
	/* Only check this afterwards, as "help" could be a valid name */
	if (0 == strcmp(name, "help")) {
		for (i = 0; list[i].name; i++) {
			printf("\t%-10s %s\n", list[i].name, list[i].description);
		}
		exit(EXIT_SUCCESS);
	}
	return undef_value;
}

static void set_option(struct xconfig_option const *option, char *arg) {
	switch (option->type) {
		case XCONFIG_BOOL:
			if (option->call)
				option->dest.func_bool(1);
			else
				*(_Bool *)option->dest.object = 1;
			break;
		case XCONFIG_BOOL0:
			if (option->call)
				option->dest.func_bool(0);
			else
				*(_Bool *)option->dest.object = 0;
			break;
		case XCONFIG_INT:
			{
				int val = strtol(arg, NULL, 0);
				if (option->call)
					option->dest.func_int(val);
				else
					*(int *)option->dest.object = val;
			}
			break;
		case XCONFIG_INT0:
			if (option->call)
				option->dest.func_int(0);
			else
				*(int *)option->dest.object = 0;
			break;
		case XCONFIG_INT1:
			if (option->call)
				option->dest.func_int(1);
			else
				*(int *)option->dest.object = 1;
			break;
		case XCONFIG_DOUBLE:
			{
				double val = strtod(arg, NULL);
				if (option->call)
					option->dest.func_double(val);
				else
					*(double *)option->dest.object = val;
			}
			break;
		case XCONFIG_STRING:
			if (option->call) {
				option->dest.func_string(arg);
			} else {
				if (*(char **)option->dest.object)
					free(*(char **)option->dest.object);
				*(char **)option->dest.object = xstrdup(arg);
			}
			break;
		case XCONFIG_STRING_LIST:
			assert(!option->call);
			*(struct slist **)option->dest.object = slist_append(*(struct slist **)option->dest.object, xstrdup(arg));
			break;
		case XCONFIG_NULL:
			if (option->call)
				option->dest.func_null();
			break;
		case XCONFIG_ENUM: {
			int val = lookup_enum(arg, (struct xconfig_enum *)option->ref, -1);
			if (option->call)
				option->dest.func_int(val);
			else
				*(int *)option->dest.object = val;
			}
			break;
		default:
			break;
	}
}

/* returns 0 if it's a value option to unset */
static int unset_option(struct xconfig_option const *option) {
	switch (option->type) {
	case XCONFIG_BOOL:
		if (option->call)
			option->dest.func_bool(0);
		else
			*(_Bool *)option->dest.object = 0;
		return 0;
	case XCONFIG_BOOL0:
		if (option->call)
			option->dest.func_bool(1);
		else
			*(_Bool *)option->dest.object = 1;
		return 0;
	case XCONFIG_INT0:
		if (option->call)
			option->dest.func_int(1);
		else
			*(int *)option->dest.object = 1;
		return 0;
	case XCONFIG_INT1:
		if (option->call)
			option->dest.func_int(0);
		else
			*(int *)option->dest.object = 0;
		return 0;
	case XCONFIG_STRING:
		if (option->call) {
			option->dest.func_string(NULL);
		} else if (*(char **)option->dest.object) {
			free(*(char **)option->dest.object);
			*(char **)option->dest.object = NULL;
		}
		return 0;
	case XCONFIG_STRING_LIST:
		assert(!option->call);
		/* providing an argument to remove here might more sense, but
		 * for now just remove the entire list: */
		slist_free_full(*(struct slist **)option->dest.object, (slist_free_func)free);
		*(struct slist **)option->dest.object = NULL;
		return 0;
	default:
		break;
	}
	return -1;
}

enum xconfig_result xconfig_set_option(struct xconfig_option const *options,
				       const char *opt, char *arg) {
	struct xconfig_option const *option = find_option(options, opt);
	if (option == NULL) {
		if (0 == strncmp(opt, "no-", 3)) {
			option = find_option(options, opt + 3);
			if (option && unset_option(option) == 0) {
				return XCONFIG_OK;
			}
		}
		LOG_ERROR("Unrecognised option `%s'\n", opt);
		return XCONFIG_BAD_OPTION;
	}
	if (option->deprecated) {
		LOG_WARN("Deprecated option `%s'\n", opt);
	}
	if (option->type == XCONFIG_BOOL ||
	    option->type == XCONFIG_BOOL0 ||
	    option->type == XCONFIG_INT0 ||
	    option->type == XCONFIG_INT1 ||
	    option->type == XCONFIG_NULL) {
		set_option(option, NULL);
		return XCONFIG_OK;
	}
	if (!arg) {
		LOG_ERROR("Missing argument to `%s'\n", opt);
		return XCONFIG_MISSING_ARG;
	}
	set_option(option, arg);
	return XCONFIG_OK;
}

/* Simple parser: one directive per line, "option argument" */
enum xconfig_result xconfig_parse_file(struct xconfig_option const *options,
		const char *filename) {
	char *line;
	FILE *cfg;
	char buf[256];
	int ret = XCONFIG_OK;

	cfg = fopen(filename, "r");
	if (cfg == NULL) return XCONFIG_FILE_ERROR;
	while ((line = fgets(buf, sizeof(buf), cfg))) {
		enum xconfig_result r = xconfig_parse_line(options, line);
		if (r != XCONFIG_OK)
			ret = r;
	}
	fclose(cfg);
	return ret;
}

enum xconfig_result xconfig_parse_line(struct xconfig_option const *options, const char *line) {
	struct xconfig_option const *option;
	char *opt, *arg;
	size_t line_len = strlen(line) + 1;
	char cline_buf[line_len];
	char *cline = cline_buf;
	strncpy(cline, line, line_len);
	cline[line_len-1] = 0;
	while (isspace((int)*cline))
		cline++;
	if (*cline == 0 || *cline == '#')
		return XCONFIG_OK;
	opt = strtok(cline, "\t\n\v\f\r =");
	if (opt == NULL) return XCONFIG_OK;
	while (*opt == '-') opt++;
	option = find_option(options, opt);
	if (option == NULL) {
		if (0 == strncmp(opt, "no-", 3)) {
			option = find_option(options, opt + 3);
			if (option && unset_option(option) == 0) {
				return XCONFIG_OK;
			}
		}
		return XCONFIG_BAD_OPTION;
	}
	if (option->deprecated) {
		LOG_WARN("Deprecated option `%s'\n", opt);
	}
	if (option->type == XCONFIG_STRING || option->type == XCONFIG_STRING_LIST) {
		/* preserve spaces */
		arg = strtok(NULL, "\n\v\f\r");
		if (arg) {
			while (isspace((int)*arg) || *arg == '=') {
				arg++;
			}
		}
	} else {
		arg = strtok(NULL, "\n\v\f\r");
		if (arg) {
			int i;
			for (i = strlen(arg)-1; i >= 0; i--) {
				if (isspace(arg[i]))
					arg[i] = 0;
			}
		}
	}
	if (option->type == XCONFIG_BOOL ||
	    option->type == XCONFIG_BOOL0 ||
	    option->type == XCONFIG_INT0 ||
	    option->type == XCONFIG_INT1 ||
	    option->type == XCONFIG_NULL) {
		set_option(option, NULL);
		return XCONFIG_OK;
	}
	if (arg == NULL) {
		LOG_ERROR("Missing argument to `%s'\n", opt);
		return XCONFIG_MISSING_ARG;
	}
	set_option(option, arg);
	return XCONFIG_OK;
}

enum xconfig_result xconfig_parse_cli(struct xconfig_option const *options,
		int argc, char **argv, int *argn) {
	int _argn = argn ? *argn : 1;
	while (_argn < argc) {
		if (argv[_argn][0] != '-') {
			break;
		}
		if (0 == strcmp("--", argv[_argn])) {
			_argn++;
			break;
		}
		char *opt = argv[_argn]+1;
		if (*opt == '-') opt++;
		struct xconfig_option const *option = find_option(options, opt);
		if (option == NULL) {
			if (0 == strncmp(opt, "no-", 3)) {
				option = find_option(options, opt + 3);
				if (option && unset_option(option) == 0) {
					_argn++;
					continue;
				}
			}
			if (argn) *argn = _argn;
			LOG_ERROR("Unrecognised option `%s'\n", opt);
			return XCONFIG_BAD_OPTION;
		}
		if (option->deprecated) {
			LOG_WARN("Deprecated option `%s'\n", opt);
		}
		if (option->type == XCONFIG_BOOL ||
		    option->type == XCONFIG_BOOL0 ||
		    option->type == XCONFIG_INT0 ||
		    option->type == XCONFIG_INT1 ||
		    option->type == XCONFIG_NULL) {
			set_option(option, NULL);
			_argn++;
			continue;
		}
		if ((_argn + 1) >= argc) {
			if (argn) *argn = _argn;
			LOG_ERROR("Missing argument to `%s'\n", opt);
			return XCONFIG_MISSING_ARG;
		}
		set_option(option, argv[_argn+1]);
		_argn += 2;
	}
	if (argn) *argn = _argn;
	return XCONFIG_OK;
}

void xconfig_shutdown(struct xconfig_option const *options) {
	for (int i = 0; options[i].type != XCONFIG_END; i++) {
		if (options[i].type == XCONFIG_STRING) {
			if (!options[i].call && *(char **)options[i].dest.object) {
				free(*(char **)options[i].dest.object);
				*(char **)options[i].dest.object = NULL;
			}
		} else if (options[i].type == XCONFIG_STRING_LIST) {
			slist_free_full(*(struct slist **)options[i].dest.object, (slist_free_func)free);
			*(struct slist **)options[i].dest.object = NULL;
		}
	}
}
